#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <getopt.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <syslog.h>
#include <ctype.h>
#include <dirent.h>
#include <time.h>
#include <inttypes.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <sys/file.h>

#define VERSION			"2.0"

#define IIO_DEVICE		"iio:device"
#define IIO_DIR			"/sys/bus/iio/devices/"
#define IIO_VIN_VOLTAGE		"in_voltage0_raw"
#define IIO_ADC_SCALE		"in_voltage_scale"
#define IIO_DEV_NAME		"3-0012"
static int iio_dev_num;

/*
 *  Regulator		-- 3.6V(+-2%) = 3.528~3.672V
 *      |
 *    Diode		-- max 0.63V voltage drop
 *      |
 * BMIC ADC Vref	-- 3.528 - 0.63 = 2.898V
 */
#define VIN_THRESH_MIN		(0U)		/* min voltage A/D conversion available in mv */
#define VIN_THRESH_MAX		(28980U)	/* max voltage A/D conversion available in mv */
						/* (equal to min voltage of ADC reference) */

#define ADC_VOLTAGE_LEN_MAX	(5) 		/* max length of in_voltage0_raw + 1 */
#define ADC_SCALE_LEN_MAX	(12)		/* max length of in_voltage_scale + 1 */

#define INTERVAL_DEFAULT	(60U) 		/* seconds */
#define INTERVAL_MIN		(0U)		/* seconds */
#define INTERVAL_MAX		(UINT_MAX)	/* seconds */

#define LOCKFILE 		"/var/run/vintrigger.lock"
static int lock = -1;

static int terminated;

#define UL_BUILD_BUG_ON_ZERO(e) (sizeof(struct { int:-!!(e); }))
#define __must_be_array(a) \
	UL_BUILD_BUG_ON_ZERO(__builtin_types_compatible_p(__typeof__(a), __typeof__(&a[0])))
#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof((arr)[0]) + __must_be_array(arr))

#define DIV_ROUND_CLOSEST(x, divisor)(			\
{							\
	typeof(divisor) __divisor = divisor;		\
	(((x) + ((__divisor) / 2)) / (__divisor));	\
}							\
)

#define __unused __attribute__((unused))

#define logging_init()							\
({									\
	openlog("vintrigger", LOG_PID | LOG_NOWAIT, LOG_DAEMON);	\
})

#define logging_close()	\
({			\
	closelog();	\
})

static void remove_exclusive_lock(void);
#define log_error(doexit, ...)			\
({						\
	syslog(LOG_ERR, __VA_ARGS__);		\
	logging_close();			\
	if (doexit) {				\
		remove_exclusive_lock();	\
		exit(EXIT_FAILURE);		\
	}					\
})

#define log_message( ...)		\
({					\
	syslog(LOG_INFO, __VA_ARGS__);	\
})


#define ALT_OVER	(1<<0)
#define ALT_UNDER	(1<<1)
#define ALT_BOTH	(ALT_OVER | ALT_UNDER)

static int alteration;
static unsigned int rising_thresh;
static unsigned int falling_thresh;
static unsigned int interval = INTERVAL_DEFAULT;

static char **params;

static ssize_t _xread(int fd, void *buf, size_t count)
{
	ssize_t ret;

	do {
		if (terminated)
			return -1;
		ret = read(fd, buf, count);
	} while (ret < 0 && errno == EINTR);

	return ret;
}

static ssize_t xread(int fd, void *buf, size_t count)
{
	ssize_t cc;
	ssize_t total;

	total = 0;

	while (count) {
		cc = _xread(fd, buf, count);

		if (cc < 0) {
			if (total)
				return total;
			return cc;
		}
		if (cc == 0)
			break;
		buf = ((char *)buf) + cc;
		total += cc;
		count -= cc;
	}

	return total;
}

static int xflock(int fd, int operation)
{
	int ret;

	do {
		ret = flock(fd, operation);
	} while (ret < 0 && errno == EINTR);

	return ret;
}

static inline unsigned int strtouint(const char *s, char **endptr, int base)
{
	uintmax_t ret = strtoumax(s, endptr, base);

	/* same as strtoul() */
	if (ret > UINT_MAX) {
		errno = ERANGE;
		return UINT_MAX;
	}

	return (unsigned int)ret;
}

static int xstrtouint(const char *str, unsigned int *val)
{
	const char *ptr;
	char *endptr;

	ptr = str;

	while (*ptr != '\0') {
		if (isspace((int)*ptr)) {
			ptr++;
			continue;
		}
		break;
	}

	if (*ptr == '\0')
		return -1;

	/* to avoid negative value */
	if (!isdigit((int)ptr[0]))
		return -1;

	errno = 0;
	*val = strtouint(ptr, &endptr, 0);

	if (errno || *endptr)
		return -1;

	return 0;
}

static int xstrtod(const char *str, double *val)
{
	const char *ptr;
	char *endptr;

	ptr = str;

	while (*ptr != '\0') {
		if (isspace((int)*ptr)) {
			ptr++;
			continue;
		}
		break;
	}

	if (*ptr == '\0')
		return -1;

	errno = 0;
	*val = strtod(ptr, &endptr);

	if (errno || *endptr)
		return -1;

	return 0;
}

static unsigned int xsleep(unsigned int seconds)
{
	unsigned int ret;

	ret = seconds;
	do {
		if (terminated)
			break;
		ret = sleep(ret);
	} while (ret);

	return ret;
}

static void get_adc_scale(double *scale, int exit_on_err)
{
	char val[ADC_SCALE_LEN_MAX];
	char path[MAXPATHLEN];
	char *c;
	int fd;
	int ret;

	snprintf(path, sizeof(path), "%s%s%d/%s", IIO_DIR, IIO_DEVICE, iio_dev_num, IIO_ADC_SCALE);
	fd = open(path, O_RDONLY);
	if (fd < 0) {
		log_error(exit_on_err, "cannot open %s", path);
		return;
	}
	ret = xread(fd, val, ADC_SCALE_LEN_MAX);
	if (ret < 0)
		log_error(exit_on_err, "cannot read %s", path);

	close(fd);

	c = strchr(val, '\n');
	if (c)
		*c = '\0';

	xstrtod(val, scale);
}

enum board_type {
	BOARDTYPE_AIOTG_G3 = 0,
	BOARDTYPE_AIOTG_G3_D2,
	BOARDTYPE_AIOTG_G3_D2_AGX3031_E00Z,
	BOARDTYPE_AIOTG_G3_W1,
	BOARDTYPE_AIOTG_G3_M1,
	BOARDTYPE_AIOTG_G3_W2,
	BOARDTYPE_AX1,
	BOARDTYPE_AX1L,
	BOARDTYPE_MAX,
};

static enum board_type x1_board_type;

#define BOARDNAME_MAX	(256)
struct board_info {
	char name[BOARDNAME_MAX];
	int r1;
	int r2;
};

static struct board_info x1_board_info[] = {
	[BOARDTYPE_AIOTG_G3] =
	{"Atmark-Techno Armadillo-IoT Gateway G3 Board", 200, 24},
	[BOARDTYPE_AIOTG_G3_D2] =
	{"Atmark-Techno Armadillo-IoT Gateway G3 D2 Board", 200, 24},
	[BOARDTYPE_AIOTG_G3_D2_AGX3031_E00Z] =
	{"Atmark-Techno Armadillo-IoT Gateway G3 D2(AGX3031-E00Z) Board", 200, 24},
	[BOARDTYPE_AIOTG_G3_W1] =
	{"Atmark-Techno Armadillo-IoT Gateway G3 W1 Board", 200, 24},
	[BOARDTYPE_AIOTG_G3_M1] =
	{"Atmark-Techno Armadillo-IoT Gateway G3 M1 Board", 200, 24},
	[BOARDTYPE_AIOTG_G3_W2] =
	{"Atmark-Techno Armadillo-IoT Gateway G3 W2 Board", 200, 24},
	[BOARDTYPE_AX1] =
	{"Atmark-Techno Armadillo-X1 Board", 200, 200},
	[BOARDTYPE_AX1L] =
	{"Atmark-Techno Armadillo-X1L Board", 200, 24},
};

/*
 *  VIN
 *   |
 *   R1
 *   |
 *   |--VOUT(to BMIC ADC)
 *   |
 *   R2
 *   |
 *  GND
 */

static int __unused vin_to_reg(int vin)
{
	struct board_info *board = &x1_board_info[x1_board_type];
	int vout;
	double scale;

	get_adc_scale(&scale, 1);
	vout = vin / scale;

	return DIV_ROUND_CLOSEST(board->r2 * vout, board->r1 + board->r2);
}

static int reg_to_vin(int reg)
{
	struct board_info *board = &x1_board_info[x1_board_type];
	int vout;
	double scale;

	get_adc_scale(&scale, 1);
	vout = reg * scale;

	return DIV_ROUND_CLOSEST((board->r1 + board->r2) * vout, board->r2);
}

static int is_valid_voltage(unsigned int voltage)
{
	if (VIN_THRESH_MAX < voltage)
		return -1;

	return 0;
}

#define IIO_MAX_NAME_LENGTH 30
static int find_iio_device_number(void)
{
	const struct dirent *ent;
	FILE *fp;
	DIR *dp;
	int num;
	char path[MAXPATHLEN];
	char name[IIO_MAX_NAME_LENGTH];

	dp = opendir(IIO_DIR);
	if (dp == NULL)
		return -1;

	for (ent = readdir(dp); ent != NULL; ent = readdir(dp)) {
		if (ent->d_name[0] == '.')
			continue;
		if (strncmp(ent->d_name, IIO_DEVICE, strlen(IIO_DEVICE)))
			continue;

		sscanf(ent->d_name, IIO_DEVICE"%d", &num);
		sprintf(path, "%s%s%d/name", IIO_DIR, IIO_DEVICE, num);
		fp = fopen(path, "r");
		if (!fp)
			continue;
		fscanf(fp, "%s", name);

		fclose(fp);
		if (!strcmp(IIO_DEV_NAME, name)) {
			closedir(dp);
			return num;
		}
	}
	closedir(dp);

	return -1;
}

static void get_vin_voltage_mv(unsigned int *voltage, int exit_on_err)
{
	char val[ADC_VOLTAGE_LEN_MAX];
	char path[MAXPATHLEN];
	char *c;
	int fd;
	int ret;

	snprintf(path, sizeof(path), "%s%s%d/%s", IIO_DIR, IIO_DEVICE, iio_dev_num, IIO_VIN_VOLTAGE);
	fd = open(path, O_RDWR);
	if (fd < 0) {
		log_error(exit_on_err, "cannot open %s", path);
		return;
	}

	ret = xread(fd, val, ADC_VOLTAGE_LEN_MAX);
	if (ret < 0)
		log_error(exit_on_err, "cannot read %s", path);

	close(fd);

	c = strchr(val, '\n');
	if (c)
		*c = '\0';

	xstrtouint(val, voltage);
	*voltage = reg_to_vin(*voltage);
}

static void wait_for_thresh(void)
{
	unsigned int voltage;

	while (1) {
		get_vin_voltage_mv(&voltage, 1);

		switch (alteration) {
		case ALT_OVER:
			if (voltage >= rising_thresh)
				return;
			break;
		case ALT_UNDER:
			if (voltage <= falling_thresh)
				return;
			break;
		case ALT_BOTH:
			if (voltage >= rising_thresh || voltage <= falling_thresh)
				return;
			break;
		default:
			log_error(1, "unsupported alteration 0x%x", alteration);
		}

		if (xsleep(interval))
			log_error(1,"cannot sleep");
	}
}

static void remove_exclusive_lock(void)
{
	if (lock < 0)
		return;

	xflock(lock, LOCK_UN | LOCK_NB);
	close(lock);
	unlink(LOCKFILE);
}

static int place_exclusive_lock(void)
{
	int ret;

	lock = open(LOCKFILE, O_RDONLY|O_CREAT|O_CLOEXEC,
		  S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH);
	if (lock < 0)
		fprintf(stderr, "cannot open %s\n", LOCKFILE);

	ret = xflock(lock, LOCK_EX | LOCK_NB);
	if (ret < 0) {
		if (errno == EWOULDBLOCK)
			fprintf(stderr, "vintrigger is already running\n");
		else
			fprintf(stderr, "cannot flock %s\n", LOCKFILE);
	}

	return ret;
}

static void term_handler(__unused int signum)
{
	terminated = 1;
}

static int setup_signal(void)
{
	int term_signals[] = {
		SIGHUP, SIGINT, SIGQUIT, SIGPIPE, SIGTERM
	};
	struct sigaction sa;
	unsigned int i;

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = term_handler;

	for (i = 0; i < ARRAY_SIZE(term_signals); i++) {
		if (sigaction(term_signals[i], &sa, NULL) < 0) {
			fprintf(stderr, "cannot set signal handler\n");
			return -1;
		}
	}

	return 0;
}

static void exec_cmd(void)
{
	int pid;
	int status;

	if (!params)
		return;

	pid = fork();
	if (pid < 0)
		log_error(1, "cannot fork");

	if (!pid) {
		execvp(params[0], params);
		log_error(1, "execvp failure");
	} else
		waitpid(pid, &status, 0);
}

static void usage(void)
{
	fprintf(stderr, "Usage: vintrigger -o|-u VOLTAGE [-i INTERVAL] [COMMAND ARGS]\n"
		"Options:\n"
		"  -o, --over=VOLTAGE\n"
		"      Execute the program COMMAND when the detected voltage is equal\n"
		"      to or over the VOLTAGE[mV].\n"
		"  -u, --under=VOLTAGE\n"
		"      Execute the program COMMAND when the detected voltage is equal\n"
		"      to or under the VOLTAGE[mV].\n"
		"VOLTAGE: Range: %u - %u\n\n"
		"  -i, --interval=INTERVAL\n"
		"      Compare with Vin to the VOLTAGE at INTERVAL second intervals.\n"
		"INTERVAL: Range: %u - %u (Default: %u)\n\n"
		"  -h, --help\n"
		"      Print usage(this message) and exit.\n"
		"  -v, --version\n"
		"      Print version information and exit.\n"
		, VIN_THRESH_MIN, VIN_THRESH_MAX, INTERVAL_MIN, INTERVAL_MAX, INTERVAL_DEFAULT);
}

static inline void version(void)
{
	fprintf(stdout, "vintrigger %s\n", VERSION);
}

static int update_board_type(void)
{
	char val[BOARDNAME_MAX];
	char path[MAXPATHLEN];
	int fd;
	int ret;
	int i;

	snprintf(path, sizeof(path), "/proc/device-tree/model");

	fd = open(path, O_RDONLY);
	if (fd < 0)
		log_error(1, "cannot open %s", path);
	ret = xread(fd, val, BOARDNAME_MAX);
	if (ret < 0)
		log_error(1, "cannot read %s", path);

	close(fd);

	for (i = 0; i < BOARDTYPE_MAX; i++) {
		if (!strncmp(val, x1_board_info[i].name, BOARDNAME_MAX)) {
			x1_board_type = i;
			return 0;
		}
	}

	return -1;
}

static int parse_args(int argc, char *argv[])
{
	struct option longopts[] = {
		{"over",required_argument, NULL, 'o' },
		{"under",required_argument, NULL, 'u' },
		{"interval",required_argument, NULL, 'i' },
		{"help",no_argument , NULL, 'h' },
		{"version",no_argument , NULL, 'v' },
		{0,0,0,0},
	};
	int c;

	while ((c = getopt_long(argc, argv, "+o:u:i:hv",longopts, NULL)) != EOF) {
		switch (c) {
		case 'o':
			alteration |= ALT_OVER;
			if (xstrtouint(optarg, &rising_thresh))
				return -1;
			if (is_valid_voltage(rising_thresh))
				return -1;
			break;
		case 'u':
			alteration |= ALT_UNDER;
			if (xstrtouint(optarg, &falling_thresh))
				return -1;
			if (is_valid_voltage(falling_thresh))
				return -1;
			break;
		case 'i':
			if (xstrtouint(optarg, &interval))
				return -1;
			break;
		case 'h':
			usage();
			return 1;
		case 'v':
			version();
			return 1;
		default:
			return -1;
		}
	}

	if (!alteration)
		return -1;

	if (optind < argc)
		params = &argv[optind];

	return 0;
}

int main(int argc, char *argv[])
{
	int ret;

	ret = parse_args(argc, argv);
	if (ret < 0) {
		usage();
		return EXIT_FAILURE;
	}
	if (ret) {
		return EXIT_SUCCESS;
	}

	ret = update_board_type();
	if (ret < 0) {
		fprintf(stderr, "unknown board type\n");
		return EXIT_FAILURE;
	}

	iio_dev_num = find_iio_device_number();
	if (iio_dev_num < 0) {
		fprintf(stderr, "cannot find %s\n", IIO_DEV_NAME);
		return EXIT_FAILURE;
	}

	ret = setup_signal();
	if (ret)
		return EXIT_FAILURE;

	ret = place_exclusive_lock();
	if (ret)
		return EXIT_FAILURE;

	logging_init();

	switch (alteration) {
	case ALT_OVER:
		log_message("waiting for an over range alert (%u mV).", rising_thresh);
		break;
	case ALT_UNDER:
		log_message("waiting for an under range alert (%u mV).", falling_thresh);
		break;
	case ALT_BOTH:
		log_message("waiting for an under range alert or a over range alert (%u-%u mV).",
			    falling_thresh, rising_thresh);
		break;
	default:
		log_error(1, "unsupported alteration 0x%x", alteration);
	}
	wait_for_thresh();
	log_message("exceeded the limit. executing command.");

	exec_cmd();

	remove_exclusive_lock();

	return EXIT_SUCCESS;
}
